#ifdef _MSC_VER	// is Microsoft compiler?
#	if _MSC_VER < 1300	// is 'old' VC 6 compiler?
#		pragma warning( disable : 4786 ) // 'identifier was truncated to '255' characters in the debug information'
#	endif // #if _MSC_VER < 1300
#endif // #ifdef _MSC_VER
#include <iostream>
#include <algorithm>
#include <map>
#include <apps/Common/exampleHelper.h>
#include <mvIMPACT_CPP/mvIMPACT_acquire.h>
#include <cassert>
#include <set>
#include <unistd.h>
#include <stdlib.h>

#define TERMINATE_APP(MSG) \
	cout << MSG << endl; \
	cout << "Press [ENTER] to end the application" << endl; \
	cin.get(); \
	return 0;

using namespace mvIMPACT::acquire;
using namespace std;

typedef map<string, Property> StringPropMap;
typedef set<int> VideoChannelContainer;

struct DeviceData;

//-----------------------------------------------------------------------------
struct CaptureParameters
//-----------------------------------------------------------------------------
{
	int maxWaitTime_ms_;
	VideoChannelContainer channels_;
	int frameCount_;
	bool boStoreImagesAsPGM_;
	bool boLogMessagePerFrame_;
	explicit CaptureParameters( int maxWaitTime_ms, int frameCount ) : maxWaitTime_ms_(maxWaitTime_ms), channels_(),
		frameCount_(frameCount), boStoreImagesAsPGM_(false), boLogMessagePerFrame_(false) {}
};

void         capture( DeviceData& devData, const CaptureParameters& captureParams );
void         captureSequence( DeviceData& devData, const CaptureParameters& captureParams );
void         createHRTCProgram( DeviceData& devData );
void         createOutputPattern( DeviceData& devData );
void         displayPropertyData( const Property& prop );
Device*      getDeviceFromUserInput( const DeviceManager& devMgr );
string       getNextParam( const string& s, string::size_type& offset );
void         loadHRTCProgram( DeviceData& devData, const string& filename );
void         modifyProperty( StringPropMap& propertyMap, const string& prop, const string& param, const string& index );
void         populatePropertyMap( StringPropMap& m, ComponentIterator it, const string& currentPath = "" );
void         printHelp( void );
void         readInputs( DeviceData& devData );
void         saveImageAsPGM( unsigned int width, unsigned int height, void* pData, const string& fileName );
unsigned int waitForInput( int maxwait_sec, int fd );

//-----------------------------------------------------------------------------
template<class _Ty>
void DeleteElement( _Ty& data )
//-----------------------------------------------------------------------------
{
	delete data;
	data = 0;
}

//-----------------------------------------------------------------------------
template<class _Elem, class _Traits, class _Ax, class _AVx>
typename std::vector<std::basic_string<_Elem, _Traits, _AVx> >::size_type split(
	/// The string to split
	const std::basic_string<_Elem,_Traits,_Ax>& str,
	/// A string defining a separator
	const std::basic_string<_Elem,_Traits,_Ax>& separator,
	/// A vector for the resulting strings
	std::vector<std::basic_string<_Elem,_Traits,_Ax>, _AVx >& v )
//-----------------------------------------------------------------------------
{
	v.clear();
	typename std::basic_string<_Elem, _Traits, _Ax>::size_type start = 0, end = 0;
	std::basic_string<_Elem, _Traits, _Ax> param, key, value;
	while( ( start = str.find_first_not_of( separator, start )) != std::basic_string<_Elem, _Traits, _Ax>::npos )
	{
		end = str.find_first_of( separator, start );
		v.push_back( ( end == std::basic_string<_Elem, _Traits, _Ax>::npos ) ? str.substr( start ) : str.substr( start, end - start ) );
		start = end;
	}
	return v.size();
}

//-----------------------------------------------------------------------------
struct DeviceData
//-----------------------------------------------------------------------------
{
	Connector connector_;
	Device* pDev_;
	FunctionInterface fi_;
	ImageRequestControl irc_;
	IOSubSystemBlueCOUGAR ioss_;
	Statistics s_;
	StringPropMap propertyMap_;
	SystemSettings ss_;
	struct VideoChannelSpecificData
	{
		explicit VideoChannelSpecificData() : propertyMap_(), setting_() {}
		StringPropMap propertyMap_;
		ComponentList setting_;
	};
	vector<VideoChannelSpecificData*> videoChannelData_;
	explicit DeviceData( Device* pDev ) : connector_(pDev), pDev_(pDev), fi_(pDev), irc_(pDev), ioss_(pDev), s_(pDev), propertyMap_(),
		ss_(pDev), videoChannelData_()
	{
		ss_.requestCount.write( ( ss_.requestCount.read( plMaxValue ) < 10 ) ? ss_.requestCount.read( plMaxValue ) : 10 );
		// get access to properties that are available once per device
		DeviceComponentLocator locator = DeviceComponentLocator(pDev, dltIOSubSystem);
		populatePropertyMap( propertyMap_, ComponentIterator(locator.searchbase_id()).firstChild() );
		locator = DeviceComponentLocator(pDev, dltRequest);
		populatePropertyMap( propertyMap_, ComponentIterator(locator.searchbase_id()).firstChild() );
		locator = DeviceComponentLocator(pDev, dltSystemSettings);
		populatePropertyMap( propertyMap_, ComponentIterator(locator.searchbase_id()).firstChild(), string("SystemSettings") );
		locator = DeviceComponentLocator(pDev, dltInfo);
		populatePropertyMap( propertyMap_, ComponentIterator(locator.searchbase_id()).firstChild(), string("Info") );
		populatePropertyMap( propertyMap_, ComponentIterator(pDev->hDev()).firstChild(), string("Device") );
		populatePropertyMap( propertyMap_, ComponentIterator(irc_.hObj()).firstChild() );
		const unsigned int VIDEO_CHANNELS = connector_.videoChannel.read( plMaxValue ) + 1;
		// For color sensors, we will NOT convert the Bayer data into a RGB image
		ImageProcessing ip(pDev);
		ip.colorProcessing.write( cpmRaw );
		for( vector<VideoChannelSpecificData*>::size_type i=0; i<VIDEO_CHANNELS; i++ )
		{
			ostringstream oss;
			oss << "VideoChannel" << i;
			VideoChannelSpecificData* p = new VideoChannelSpecificData();
			fi_.createSetting( oss.str(), "Base", &p->setting_ );
			DeviceComponentLocator locator(pDev, dltSetting, oss.str());
			// get access to the properties that are available once per videochannel
			populatePropertyMap( p->propertyMap_, ComponentIterator(locator.searchbase_id()).firstChild() );
			StringPropMap::const_iterator it = p->propertyMap_.find( "Connector/VideoChannel" );
			assert( ( it != propertyMap_.end() ) && "Could not detect property 'Connector/VideoChannel'" );
			if( it != propertyMap_.end() )
			{
				PropertyI videoChannel(it->second.hObj());
				// prepare this setting to use a certain video channel
				// thus each setting will operate with the video channel specific parameters then
				videoChannel.write( i );
			}
			videoChannelData_.push_back( p );
		}
	}
	~DeviceData()
	{
		for_each( videoChannelData_.begin(), videoChannelData_.end(), ptr_fun(DeleteElement<VideoChannelSpecificData*>) );
	}
};

//-----------------------------------------------------------------------------
class DisplayPropertyMap : public unary_function<DeviceData::VideoChannelSpecificData*, void>
//-----------------------------------------------------------------------------
{
public:
	void operator()( const DeviceData::VideoChannelSpecificData* p ) const
	{
		if( p )
		{
			for_each( p->propertyMap_.begin(), p->propertyMap_.end(), DisplayProperty() );
		}
	}
};

//-----------------------------------------------------------------------------
void capture( DeviceData& devData, const CaptureParameters& captureParams )
//-----------------------------------------------------------------------------
{
	if( captureParams.channels_.empty() )
	{
		cout << "ERROR: No video channels specified. Can't capture image data" << endl;
		return;
	}

	cout << "Starting to capture " << captureParams.frameCount_ << " frame(s) from " << captureParams.channels_.size() << " (different) video channel(s)" << endl;

	int framesRequested = 0;
	int result = DMR_NO_ERROR;
	// prefill the capture queue
	VideoChannelContainer::const_iterator it = captureParams.channels_.begin();
	VideoChannelContainer::const_iterator itEND = captureParams.channels_.end();
	while( ( result == DMR_NO_ERROR ) && ( framesRequested < captureParams.frameCount_ ) && ( it != itEND ) )
	{
		// select setting for the video channel to capture from
		devData.irc_.setting.write( devData.videoChannelData_[*it]->setting_.hObj() );
		if( ( result = devData.fi_.imageRequestSingle( &devData.irc_ ) ) == DMR_NO_ERROR )
		{
			++it;
			if( it == itEND )
			{
				it = captureParams.channels_.begin();
			}
			++framesRequested;
		}
	}

	int framesDelivered = 0;
	bool boRun = true;
	while( boRun )
	{
		// wait for the image send to the default capture queue
		int requestNr = devData.fi_.imageRequestWaitFor( captureParams.maxWaitTime_ms_ );
		// check if the image has been captured without any problems
		if( devData.fi_.isRequestNrValid( requestNr ) )
		{
			++framesDelivered;
			const Request* pRequest = devData.fi_.getRequest(requestNr);
			if( devData.fi_.isRequestOK( pRequest ) )
			{
				// everything went well.
				if( ( pRequest->imageWidth.read() == 0 ) || ( pRequest->imageHeight.read() == 0 ) )
				{
					cout << "ERROR! There is an image marked as OK which isn't" << endl;
				}
				else
				{
					if( captureParams.boLogMessagePerFrame_ )
					{
						cout << "Image " << pRequest->infoFrameNr.read() << "(Id: " << pRequest->infoFrameID.read()
							<< ", timestamp(us): " << pRequest->infoTimeStamp_us.read() << ", expose_us: " << pRequest->infoExposeTime_us.read()
							<< ", IOStates: " << pRequest->infoIOStatesAtExposureStart.readS() << "(s), " << pRequest->infoIOStatesAtExposureEnd.readS() << "(e)"
							<< ", " << pRequest->imageWidth.read() << "x" << pRequest->imageHeight.read()
							<< ", " << pRequest->imagePixelFormat.readS() << ") captured from channel "
							<< pRequest->infoVideoChannel.read() << ". Frames delivered: " << framesDelivered << endl;
					}
					if( captureParams.boStoreImagesAsPGM_ )
					{
						ostringstream oss;
						oss << devData.pDev_->serial.read() << "_channel" << pRequest->infoVideoChannel.read() << "_" << pRequest->imageWidth.read()
							<< "x" << pRequest->imageHeight.read() << "_" << pRequest->imagePixelFormat.readS() << "_Frame" << framesDelivered <<".pgm";
						saveImageAsPGM( static_cast<unsigned int>(pRequest->imageWidth.read()), static_cast<unsigned int>(pRequest->imageHeight.read()), pRequest->imageData.read(), oss.str() );
					}
				}
			}
			else
			{
				cout << "A request from channel " << pRequest->infoVideoChannel.read() << " has been returned, but the acqusition was not successful. Reason: " << pRequest->requestResult.readS() << endl;
			}
			// ... unlock the request again, so that the driver can use it again
			devData.fi_.imageRequestUnlock( requestNr );
			if( framesRequested < captureParams.frameCount_ )
			{
				// request a new image from the channel that returned this request
				devData.irc_.setting.write( devData.videoChannelData_[pRequest->infoVideoChannel.read()]->setting_.hObj() );
				result = devData.fi_.imageRequestSingle( &devData.irc_ );
				if( result != DMR_NO_ERROR )
				{
					cout << "Calling imageRequestSingle for setting " << devData.irc_.setting.readS() << " failed: " << ImpactAcquireException::getErrorCodeAsString( result ) << endl;
				}
				++framesRequested;
			}
			boRun = ( framesDelivered >= captureParams.frameCount_ ) ? false : true;
		}
		else
		{
			cout << "The acquisition failed: " << ImpactAcquireException::getErrorCodeAsString( requestNr ) << endl;
		}

		if( boRun )
		{
			boRun = ( waitForInput( 0, STDIN_FILENO ) == 0 ) ? true : false; // break by STDIN
		}
	}

	cout << devData.s_.framesPerSecond.name() << ": " << devData.s_.framesPerSecond.read() << endl;
	// in any case clear the queue to have consistent behaviour the next time this function gets called
	// as otherwise an image not ready yet might be returned directly when this function gets called the next time
	it = captureParams.channels_.begin();
	while( it != itEND )
	{
		devData.fi_.imageRequestReset( 0, *it );
		++it;
	}
}

//-----------------------------------------------------------------------------
void captureSequence( DeviceData& devData, const CaptureParameters& captureParams )
//-----------------------------------------------------------------------------
{
	if( captureParams.channels_.empty() )
	{
		cout << "ERROR: No video channels specified. Can't capture image data" << endl;
		return;
	}

	cout << "Starting to capture from " << captureParams.channels_.size() << " (different) video channel(s)" << endl;

	int framesRequested = 0;
	int result = DMR_NO_ERROR;
	// request as many images as possible
	VideoChannelContainer::const_iterator it = captureParams.channels_.begin();
	VideoChannelContainer::const_iterator itEND = captureParams.channels_.end();
	while( ( result == DMR_NO_ERROR ) && ( it != itEND ) )
	{
		// select setting for the video channel to capture from
		devData.irc_.setting.write( devData.videoChannelData_[*it]->setting_.hObj() );
		if( ( result = devData.fi_.imageRequestSingle( &devData.irc_ ) ) == DMR_NO_ERROR )
		{
			++it;
			if( it == itEND )
			{
				it = captureParams.channels_.begin();
			}
			++framesRequested;
		}
	}

	int framesDelivered = 0;
	bool boRun = true;
	while( boRun )
	{
		// wait for the image send to the default capture queue
		int requestNr = devData.fi_.imageRequestWaitFor( captureParams.maxWaitTime_ms_ );
		// check if the image has been captured without any problems
		if( devData.fi_.isRequestNrValid( requestNr ) )
		{
			++framesDelivered;
			boRun = ( framesRequested > framesDelivered ) ? true : false;
		}
		else
		{
			cout << "The acquisition failed: " << ImpactAcquireException::getErrorCodeAsString( requestNr ) << endl;
		}

		if( boRun )
		{
			boRun = ( waitForInput( 0, STDIN_FILENO ) == 0 ) ? true : false; // break by STDIN
		}
	}

	cout << devData.s_.framesPerSecond.name() << ": " << devData.s_.framesPerSecond.read() << endl;

	for( int i=0; i<framesDelivered; i++ )
	{
		const Request* pRequest = devData.fi_.getRequest(i);
		if( devData.fi_.isRequestOK( pRequest ) )
		{
			// everything went well.
			if( ( pRequest->imageWidth.read() == 0 ) || ( pRequest->imageHeight.read() == 0 ) )
			{
				cout << "ERROR! There is an image marked as OK which isn't" << endl;
			}
			else if( captureParams.boStoreImagesAsPGM_ )
			{
				ostringstream oss;
				oss << devData.pDev_->serial.read() << "_channel" << pRequest->infoVideoChannel.read() << "_Id: " << pRequest->infoFrameID.read() << "_" << pRequest->imageWidth.read()
					<< "x" << pRequest->imageHeight.read() << "_" << pRequest->imagePixelFormat.readS() << "_Frame" << i <<".pgm";
				saveImageAsPGM( static_cast<unsigned int>(pRequest->imageWidth.read()), static_cast<unsigned int>(pRequest->imageHeight.read()), pRequest->imageData.read(), oss.str() );
			}
			devData.fi_.imageRequestUnlock( i );
		}
	}

	// in any case clear the queue to have consistent behaviour the next time this function gets called
	// as otherwise an image not ready yet might be returned directly when this function gets called the next time
	it = captureParams.channels_.begin();
	while( it != itEND )
	{
		devData.fi_.imageRequestReset( 0, *it );
		++it;
	}
}

//-----------------------------------------------------------------------------
void createHRTCProgram( DeviceData& devData )
//-----------------------------------------------------------------------------
{
	unsigned int HRTCCount = devData.ioss_.RTCtrProgramCount();
	if( HRTCCount == 0 )
	{
		cout << "This device doesn't seem to support HRTCs. A firmware update and/or interface driver update might fix this" << endl;
		return;
	}

	RTCtrProgram* pRTCtrlprogram = devData.ioss_.getRTCtrProgram( 0 );
	if( !pRTCtrlprogram )
	{
		cout << "ERROR! The device claims to support HRTCs but does not report an exisisting controller?!" << endl;
		return;
	}

	pRTCtrlprogram->setProgramSize( 22 );

	// 0. switch on some outputs
	int progStep = 0;
	RTCtrProgramStep* pRTCtrlStep = 0;
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgSetDigout );
	assert( !( pRTCtrlStep->digitalOutputs.valCount() < 4 ) && "Expecting at least 4 digital outputs" );
	pRTCtrlStep->digitalOutputs.write( digioOn, 1 );
	pRTCtrlStep->digitalOutputs.write( digioOn, 2 );

	// 1. wait
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgWaitClocks );
	pRTCtrlStep->clocks_us.write( 10000 );

	// 2. switch off these outputs and turn on some others
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgSetDigout );
	pRTCtrlStep->digitalOutputs.write( digioOn, 0 );
	pRTCtrlStep->digitalOutputs.write( digioOff, 1 );
	pRTCtrlStep->digitalOutputs.write( digioOff, 2 );
	pRTCtrlStep->digitalOutputs.write( digioOn, 3 );

	// 3. wait
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgWaitClocks );
	pRTCtrlStep->clocks_us.write( 20000 );

	// 4. switch off all outputs
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgSetDigout );
	pRTCtrlStep->digitalOutputs.write( digioOff, 0 );
	pRTCtrlStep->digitalOutputs.write( digioOff, 1 );
	pRTCtrlStep->digitalOutputs.write( digioOff, 2 );
	pRTCtrlStep->digitalOutputs.write( digioOff, 3 );

	// 5. wait
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgWaitClocks );
	pRTCtrlStep->clocks_us.write( 10000 );

	int loopCnt = 4;
	// 6. define a loop with 'loopCnt' runs
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgRegisterSet );
	pRTCtrlStep->controllerRegister.write( 0 );
	pRTCtrlStep->registerValue.write( loopCnt );
	{
		// this is the loop
		int loopAddress = progStep;

		// 7. wait
		pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
		pRTCtrlStep->opCode.write( rtctrlProgWaitClocks );
		pRTCtrlStep->clocks_us.write( 5000 );

		// 8. switch on an output
		pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
		pRTCtrlStep->opCode.write( rtctrlProgSetDigout );
		pRTCtrlStep->digitalOutputs.write( digioOn, 2 );

		// 9. wait
		pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
		pRTCtrlStep->opCode.write( rtctrlProgWaitClocks );
		pRTCtrlStep->clocks_us.write( 5000 );

		// 10. switch it off again
		pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
		pRTCtrlStep->opCode.write( rtctrlProgSetDigout );
		pRTCtrlStep->digitalOutputs.write( digioOff, 2 );

		// 11. decrement the register value
		pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
		pRTCtrlStep->opCode.write( rtctrlProgRegisterSub );
		pRTCtrlStep->controllerRegister.write( 0 );
		pRTCtrlStep->registerValue.write( 1 );

		// 12. jump back to the start of the loop until the register value reached 0
		pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
		pRTCtrlStep->opCode.write( rtctrlProgJumpLocOnNotZero );
		pRTCtrlStep->controllerRegister.write( 0 );
		pRTCtrlStep->address.write( loopAddress );
	}

	// 13. check the jumpOnZero and the registerAdd cmd
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgRegisterSet );
	pRTCtrlStep->controllerRegister.write( 1 );
	pRTCtrlStep->registerValue.write( 1 );

	// 14. jump over the next lines in case of a bug
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgJumpLocOnZero );
	pRTCtrlStep->controllerRegister.write( 1 );
	pRTCtrlStep->address.write( 18 );

	// 15. wait
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgWaitClocks );
	pRTCtrlStep->clocks_us.write( 1500 );

	// 16. switch on an output
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgSetDigout );
	pRTCtrlStep->digitalOutputs.write( digioOn, 3 );

	// 17. wait
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgWaitClocks );
	pRTCtrlStep->clocks_us.write( 1500 );

	// 18. switch it off again
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgSetDigout );
	pRTCtrlStep->digitalOutputs.write( digioOff, 3 );

	// 19. wait
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgWaitClocks );
	pRTCtrlStep->clocks_us.write( 2500 );

	// 20. NOP
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgNop );

	// 21. start from the beginning again
	pRTCtrlStep = pRTCtrlprogram->programStep( progStep++ );
	pRTCtrlStep->opCode.write( rtctrlProgJumpLoc );
	pRTCtrlStep->address.write( 0 );

	// start the program
	pRTCtrlprogram->mode.write( rtctrlModeRun );

	cout << "Now a blinking test pattern should be visible on the digital outputs" << endl;
	cout << "Press [ENTER] to stop the HRTC test" << endl;
	do {} while( ( waitForInput( 100, STDIN_FILENO ) == 0 ) );

	pRTCtrlprogram->mode.write( rtctrlModeStop );
}

//-----------------------------------------------------------------------------
void loadHRTCProgram( DeviceData& devData, const string& filename )
//-----------------------------------------------------------------------------
{
	unsigned int HRTCCount = devData.ioss_.RTCtrProgramCount();
	if( HRTCCount == 0 )
	{
		cout << "This device doesn't seem to support HRTCs. A firmware update and/or interface driver update might fix this" << endl;
		return;
	}

	RTCtrProgram* pRTCtrlProgram = devData.ioss_.getRTCtrProgram( 0 );
	if( !pRTCtrlProgram )
	{
		cout << "ERROR! The device claims to support HRTCs but does not report an exisisting controller?!" << endl;
		return;
	}

	pRTCtrlProgram->filename.write( filename );
	int result = pRTCtrlProgram->load();
	if( result != DMR_NO_ERROR )
	{
		cout << "Could not load HRTC program '" << filename << "'. Result: " << pRTCtrlProgram->programState.read() << endl;
		return;
	}

	// start the program
	pRTCtrlProgram->mode.write( rtctrlModeRun );
	cout << pRTCtrlProgram->programState.read() << endl;
	cout << "The HRTC program should run if no error have been reported before. Make sure that 'Camera/TriggerSource' is set the" << endl
		<< " 'RTCtrl' and select the appropriate value for 'Camera/TriggerMode'." << endl;
}

//-----------------------------------------------------------------------------
void createOutputPattern( DeviceData& devData )
//-----------------------------------------------------------------------------
{
	cout << "Press [ENTER] to stop the output test" << endl;
	const unsigned int OUTPUT_CNT = devData.ioss_.getOutputCount();
	while( ( waitForInput( 0, STDIN_FILENO ) == 0 ) ? true : false )
	{
		for( unsigned int i=0; i<OUTPUT_CNT; i++ )
		{
			devData.ioss_.output( i )->set();
			usleep( 250000 );
			devData.ioss_.output( i )->reset();
		}
		for( unsigned int i=0; i<OUTPUT_CNT; i++ )
		{
			devData.ioss_.output( i )->set();
		}
		usleep( 1000000 );
		for( unsigned int i=0; i<OUTPUT_CNT; i++ )
		{
			devData.ioss_.output( i )->reset();
		}
	}
}

//-----------------------------------------------------------------------------
string getNextParam( const string& s, string::size_type& offset )
//-----------------------------------------------------------------------------
{
	offset = s.find_first_not_of( " \t", offset );
	if( offset == string::npos )
	{
		return string("");
	}

	string::size_type end = s.find_first_of( " \t", offset + 1 );
	string result(s.substr( offset, ( end == string::npos ) ? string:: npos : end - offset ));
	offset = end;
	return result;
}

//-----------------------------------------------------------------------------
void modifyProperty( StringPropMap& propertyMap, const string& prop, const string& param, const string& index )
//-----------------------------------------------------------------------------
{
	StringPropMap::const_iterator it = propertyMap.find( prop );
	if( it == propertyMap.end() )
	{
		cout << "unknown property: " << prop << "." << endl;
	}
	else
	{
		if( param.empty() )
		{
			displayPropertyData( it->second );
		}
		modifyPropertyValue( it->second, param, index );
	}
}

//-----------------------------------------------------------------------------
void populatePropertyMap( StringPropMap& m, ComponentIterator it, const string& currentPath /* = "" */ )
//-----------------------------------------------------------------------------
{
	while( it.isValid() )
	{
		string fullName( currentPath );
		if( fullName != "" )
		{
			fullName += "/";
		}
		fullName += it.name();
		if( it.isList() )
		{
			populatePropertyMap( m, it.firstChild(), fullName );
		}
		else if( it.isProp() )
		{
			m.insert( make_pair( fullName, Property( it ) ) );
		}
		++it;
		// method objects will be ignored...
	}
}

//-----------------------------------------------------------------------------
void printHelp( void )
//-----------------------------------------------------------------------------
{
	cout << " Available commands / usage:" << endl
		<< " ------------------------------------" << endl
		<< " quit: terminates the sample" << endl
		<< endl
		<< " cap <arg>: acquire images" << endl
		<< "   Arguments:" << endl
		<< "     -h<nr> capture from the video channel specified by number" << endl
		<< "     -n<nr> capture <nr> images from the video channel specified" << endl
		<< "     -s     store all captured images as PGM files in the current directory" << endl
		<< "     -l     write a log message for each image captured to stdout" << endl
		<< "     Example: cap -h0 -h1 -n5 -s// will capture 5 frames from channel 0 and channel 1 and will store the images in the current directory as PGM files" << endl
		<< "     Please note that the number of frames to capture specifies the total number of images not the number of images per video channel" << endl
		<< " caps <arg>: acquire images" << endl
		<< "   Arguments:" << endl
		<< "     -h<nr> capture from the video channel specified by number" << endl
		<< "     -s     store all captured images as PGM files in the current directory" << endl
		<< "     Example: caps -h0 -h1 -s // will capture as many frames as there are request objects from channel 0 and channel 1 and will store the images in the current directory as PGM files after all images have been captured" << endl
		<< "     Please note that the number of frames to capture specifies the total number of images not the number of images per video channel" << endl
		<< " list <arg>: displays properties available for this device" << endl
		<< "   Arguments:" << endl
		<< "     -f     displays all parameters" << endl
		<< "     -d     displays the device specific parameters only" << endl
		<< "     -h<nr> displays the parameters associated with the video channel specified by number" << endl
		<< endl
		<< " mod <arg> <fullName>: modifies the property specified in <fullName>." << endl
		<< "   Arguments:" << endl
		<< "     -d     specifies a device specific parameter" << endl
		<< "     -h<nr> specifies a parameter associated with the video channel specified by <nr>" << endl
		<< "     Example: mod -h0 Camera/Aoi/W" << endl
		<< endl
		<< " outputs: checks the digital outputs by creating a periodical pattern" << endl
		<< endl
		<< " inputs: returns the current state of the digital inputs" << endl
		<< endl
		<< " HRTC: checks the operation of the HRTC by creating a HRTC program that will be executed continuously" << endl
		<< endl
		<< " loadHRTCProgram: Tries to load an start a HRTC program from the specified location" << endl
		<< endl
		<< " help: displays this help text" << endl
		<< endl
		<< " timeout: set a new timeout(in ms) used as a max. timeout to wait for an image" << endl
		<< endl;
}

//-----------------------------------------------------------------------------
void readInputs( DeviceData& devData )
//-----------------------------------------------------------------------------
{
	unsigned int inputPattern1 = devData.ioss_.readInputRegister();
	unsigned int inputPattern2 = 0;
	const unsigned int INPUT_COUNT = devData.ioss_.getInputCount();
	for( unsigned int i=0; i<INPUT_COUNT; i++ )
	{
		inputPattern2 |= devData.ioss_.input( i )->get() << i;
	}
	cout << "Result of readInputRegister: " << inputPattern1 << endl;
	cout << "Result of collecting the data of each input: " << inputPattern2 << endl;
}

//-----------------------------------------------------------------------------
void saveImageAsPGM( unsigned int width, unsigned int height, void* pData, const string& fileName )
//-----------------------------------------------------------------------------
{
	if( pData )
	{
		unsigned int iImageFileSize = width * height;
		unsigned char* pImagePtr = static_cast<unsigned char*>(pData);
		FILE* fp = fopen( fileName.c_str(), "wb" );
		if( fp != 0 )
		{
			ostringstream oss;
			oss << "P5\n# Created by MATRIX VISION GmbH" << endl
				<< width << endl
				<< height << endl
				<< "255" << endl;
			fwrite( oss.str().c_str(), 1, oss.str().length(), fp );
			size_t retCount = 0;
			if( ( retCount = fwrite( pImagePtr, 1, iImageFileSize, fp ) ) != iImageFileSize )
			{
				cout << string(__FUNCTION__) << ": only " << retCount << " bytes written to file " << fileName << "s! Expected: " << iImageFileSize << endl;
			}
			fclose( fp );
		}
		else
		{
			cout << string(__FUNCTION__) << ": Cannot create file " << fileName << endl;
		}
	}
}

//-----------------------------------------------------------------------------
// returns 0 if timeout, else 1
unsigned waitForInput( int maxwait_sec, int fd )
//-----------------------------------------------------------------------------
{
	fd_set rfds;
	struct timeval tv;

	FD_ZERO(&rfds);
	FD_SET(fd, &rfds);

	tv.tv_sec = maxwait_sec ;
	tv.tv_usec = 0;

	return select(fd+1, &rfds, NULL, NULL, &tv);
}

//-----------------------------------------------------------------------------
bool isDeviceSupportedBySample( const Device* const pDev )
//-----------------------------------------------------------------------------
{
	return ( match( pDev->product.read(), string("mvBlueLYNX-M7"), '*' ) == 0 );
}

//-----------------------------------------------------------------------------
int main( int argc, char* argv[] )
//-----------------------------------------------------------------------------
{
	DeviceManager devMgr;
	Device* pDev = 0;

	string cmdLine;
	for( int i=1; i<argc; i++ )
	{
		cmdLine.append( string(argv[i]) );
		cmdLine.append( string(" ") );
	}

	vector<string> cmdList;
	split( cmdLine, string(";"), cmdList );

	if( !cmdList.empty() )
	{
		cout << "Trying to obtain device pointer by serial using arg " << cmdList[0] << endl;
		pDev = devMgr.getDeviceBySerial( cmdList[0] );
		cmdList.erase( cmdList.begin() );
	}
	else
	{
		pDev = getDeviceFromUserInput( devMgr, isDeviceSupportedBySample );
	}

	if( pDev == 0 )
	{
		TERMINATE_APP( "Invalid device pointer. Unable to continue!" )
	}

	if( pDev->product.read() != "mvBlueLYNX-M7" )
	{
		TERMINATE_APP( "This sample is currently meant for mvBlueLYNX-M7 devices only!" )
	}

	cout << "Initialising the device. This might take some time..." << endl;
	try
	{
		pDev->open();
	}
	catch( const ImpactAcquireException& e )
	{
		// this e.g. might happen if the same device is already opened in another process...
		ostringstream oss;
		oss << "An error occurred while opening the device(error code: " << e.getErrorCode() << ").";
		TERMINATE_APP( oss.str() )
	}

	DeviceData devData(pDev);
	int timeout_ms = 500;
	printHelp();
	bool boRun = true;
	while( boRun )
	{
		if( !cmdList.empty() )
		{
			cmdLine = *cmdList.begin();
			cmdList.erase( cmdList.begin() );
		}
		else
		{
			cout << "Enter the command line followed by [ENTER]: ";
			getline( cin, cmdLine );
		}
		cout << "Processing command line '" << cmdLine << "'" << endl;

		string::size_type offset = 0;
		while( offset != string::npos )
		{
			string cmd(getNextParam( cmdLine, offset ));
			if( cmd.empty() )
			{
				break;
			}

			if( cmd == "cap" )
			{
				vector<string> args;
				string arg;
				while( ( arg = getNextParam( cmdLine, offset ) ) != string("") )
				{
					args.push_back( arg );
				}
				CaptureParameters captureParams(timeout_ms, 1);
				vector<string>::size_type vSize = args.size();
				for( vector<string>::size_type i=0; i<vSize; i++ )
				{
					if( args[i] == "-s" )
					{
						captureParams.boStoreImagesAsPGM_ = true;
					}
					else if( args[i] == "-l" )
					{
						captureParams.boLogMessagePerFrame_ = true;
					}
					else if( ( args[i].length() > 2 ) && isdigit(  args[i][2] ) )
					{
						string indexAsString( args[i].substr( 2 ));
						istringstream iss(indexAsString);
						int index = 0;
						iss >> index;
						if( args[i].find( "-h" ) == 0 )
						{
							if( ( index < 0 ) || ( index > devData.connector_.videoChannel.read( plMaxValue ) ) )
							{
								cout << "Invalid video channel index (" << index << ", arg: " <<  args[i] << ") passed to 'cap' command. SKIPPED!" << endl;
								continue;
							}
							captureParams.channels_.insert( index );
						}
						else if( args[i].find( "-n" ) == 0 )
						{
							if( index < 0 )
							{
								cout << "Invalid frame count (" << index << ", arg: " <<  args[i] << ") passed to 'cap' command. SKIPPED!" << endl;
								continue;
							}
							captureParams.frameCount_ = index;
						}
						else
						{
							cout << "Invalid argument (" << index << ", arg: " <<  args[i] << ") passed to 'cap' command. SKIPPED!" << endl;
						}
					}
					else
					{
						cout << "Invalid argument ( arg: " <<  args[i] << ") passed to 'cap' command. SKIPPED!" << endl;
					}
				}
				capture( devData, captureParams );
			}
			else if( cmd == "caps" )
			{
				vector<string> args;
				string arg;
				while( ( arg = getNextParam( cmdLine, offset ) ) != string("") )
				{
					args.push_back( arg );
				}
				CaptureParameters captureParams(timeout_ms, 1);
				vector<string>::size_type vSize = args.size();
				for( vector<string>::size_type i=0; i<vSize; i++ )
				{
					if( args[i] == "-s" )
					{
						captureParams.boStoreImagesAsPGM_ = true;
					}
					else if( ( args[i].length() > 2 ) && isdigit(  args[i][2] ) )
					{
						string indexAsString( args[i].substr( 2 ));
						istringstream iss(indexAsString);
						int index = 0;
						iss >> index;
						if( args[i].find( "-h" ) == 0 )
						{
							if( ( index < 0 ) || ( index > devData.connector_.videoChannel.read( plMaxValue ) ) )
							{
								cout << "Invalid video channel index (" << index << ", arg: " <<  args[i] << ") passed to 'cap' command. SKIPPED!" << endl;
								continue;
							}
							captureParams.channels_.insert( index );
						}
						else
						{
							cout << "Invalid argument (" << index << ", arg: " <<  args[i] << ") passed to 'cap' command. SKIPPED!" << endl;
						}
					}
					else
					{
						cout << "Invalid argument ( arg: " <<  args[i] << ") passed to 'cap' command. SKIPPED!" << endl;
					}
				}
				captureSequence( devData, captureParams );
			}
			else if( cmd == "list" )
			{
				string arg(getNextParam( cmdLine, offset ));
				if( arg == "-f" )
				{
					for_each( devData.propertyMap_.begin(), devData.propertyMap_.end(), DisplayProperty() );
					for_each( devData.videoChannelData_.begin(), devData.videoChannelData_.end(), DisplayPropertyMap() );
				}
				else if( arg == "-d" )
				{
					for_each( devData.propertyMap_.begin(), devData.propertyMap_.end(), DisplayProperty() );
				}
				else if( ( arg.find( "-h" ) == 0 ) && ( arg.length() > 2 ) && isdigit( arg[2] ) )
				{
					string indexAsString(arg.substr( 2 ));
					istringstream iss(indexAsString);
					int index = 0;
					iss >> index;
					if( ( index < 0 ) || ( index > devData.connector_.videoChannel.read( plMaxValue ) ) )
					{
						cout << "Invalid video channel index (" << index << ", arg: " << arg << ") passed to 'list' command. SKIPPED!" << endl;
						continue;
					}
					for_each( devData.videoChannelData_[index]->propertyMap_.begin(), devData.videoChannelData_[index]->propertyMap_.end(), DisplayProperty() );
				}
				else
				{
					cout << "Invalid argument (" << arg << ") passed to 'list' command. SKIPPED!" << endl;
				}
			}
			else if( cmd == "mod" )
			{
				vector<string> args;
				string arg;
				while( ( arg = getNextParam( cmdLine, offset ) ) != string("") )
				{
					args.push_back( arg );
				}
				if( args.size() >= 2 )
				{
					if( args[0] == "-d" )
					{
						modifyProperty( devData.propertyMap_, args[1], ( args.size() > 2 ) ? args[2] : string(""), ( args.size() > 3 ) ? args[3] : string("") );
					}
					else if( ( args[0].find( "-h" ) == 0 ) && ( args[0].length() > 2 ) && isdigit( args[0][2] ) )
					{
						string indexAsString(args[0].substr( 2 ));
						istringstream iss(indexAsString);
						int index = 0;
						iss >> index;
						if( ( index < 0 ) || ( index > devData.connector_.videoChannel.read( plMaxValue ) ) )
						{
							cout << "Invalid video channel index (" << index << ", arg: " << args[0] << ") passed to 'mod' command. SKIPPED!" << endl;
							continue;
						}
						modifyProperty( devData.videoChannelData_[index]->propertyMap_, args[1], ( args.size() > 2 ) ? args[2] : string(""), ( args.size() > 3 ) ? args[3] : string("") );
					}
					else
					{
						cout << "Invalid argument (" << args[0] << ") passed to 'mod' command. SKIPPED!" << endl;
					}
				}
				else
				{
					cout << "Not enough parameters passed to 'mod' command. SKIPPED!" << endl;
				}
			}
			else if( cmd == "help" )
			{
				printHelp();
			}
			else if( cmd == "outputs" )
			{
				createOutputPattern( devData );
			}
			else if( cmd == "inputs" )
			{
				readInputs( devData );
			}
			else if( cmd == "HRTC" )
			{
				createHRTCProgram( devData );
			}
			else if( cmd == "loadHRTCProgram" )
			{
				loadHRTCProgram( devData, getNextParam( cmdLine, offset ) );
			}
			else if( cmd == "quit" )
			{
				boRun = false;
				break;
			}
			else if( cmd == "timeout" )
			{
				cout << "Enter the new timeout to be passed to the imageRequestWaitFor function: ";
				cin >> timeout_ms;
			}
			else
			{
				cout << "unknown command: " << cmd << ". SKIPPED!" << endl;
				printHelp();
			}
		}
		cmdLine.clear();
	}

	return 0;
}
